@@ -25,9 +25,9 @@ compute.gene.cdf <- function(expr, Gaussk=TRUE, kernel=TRUE,
                 gene.cdf <- .kcdfvals_sparse_to_dense(expr, Gaussk, verbose)
         } else if (is(expr, "SVT_SparseArray")) {
             if (sparse)
-                gene.cdf <- .kcdfvals_svt_to_svt(expr, verbose)
+                gene.cdf <- .kcdfvals_svt_to_svt(expr, Gaussk, verbose)
             else
-                gene.cdf <- .kcdfvals_svt_to_dense(expr, verbose)
+                gene.cdf <- .kcdfvals_svt_to_dense(expr, Gaussk, verbose)
         } else if (is(expr, "DelayedMatrix")) {
             if (sparse)
                 gene.cdf <- .kcdfvals_sparseh5_to_sparseh5(expr, Gaussk=Gaussk,
@@ -58,7 +58,8 @@ compute.gene.cdf <- function(expr, Gaussk=TRUE, kernel=TRUE,
                       verbose)
             gene.cdf <- t(matrix(A, n.test.samples, n.genes))
         } else
-            cli_abort("x"=sprintf("Matrix class %s cannot be handled yet.", class(expr)))
+            cli_abort(c("x"=sprintf("Matrix class %s cannot be handled yet.",
+                                    class(expr))))
     } else { ## direct ECDF estimation
         if (is(expr, "dgCMatrix")) {
             if (sparse)
@@ -88,8 +89,8 @@ compute.gene.cdf <- function(expr, Gaussk=TRUE, kernel=TRUE,
             else
                 gene.cdf <- .ecdfvals_dense_to_dense(expr, verbose)
         } else
-            cli_abort("x"=sprintf("Input container class %s cannot be handled yet.",
-                         class(expr)))
+            cli_abort(c("x"=sprintf("Input container class %s cannot be handled yet.",
+                                    class(expr))))
     }
 
     return(gene.cdf)	
@@ -221,8 +222,19 @@ zorder_rankstat <- function(z, p) {
 #' @param verbose Gives information about each calculation step. Default: `TRUE`.
 #'
 #' @param BPPARAM An object of class `BiocParallelParam` specifying parameters
-#'   related to the parallel execution of some of the tasks and calculations
-#'   within this function.
+#' related to the parallel execution of some of the tasks and calculations
+#' within this function.
+#'
+#' @param maxmem A vector of length 1 either specifying a number in bytes, or
+#' a character string with either the word `auto` (default), or a number
+#' followed by a suffix indicating kilobytes (K), megabytes (M), gigabytes (G)
+#' or terabytes (T), which GSVA will use to attempt bounding the maximum amount
+#' of main memory used across all threads of execution to that given quantity.
+#' By default `maxmem="auto"`, indicating that the maximum memory will be the
+#' 90% of the total main memory, as calculated by [`Sys.meminfo()`][memuse::Sys.meminfo].
+#' To avoid setting any bound on the maximum memory, please use `maxmem=Inf`.
+#' Note that the amount of main memory used in an R session or script may depend
+#' on other commands and packages used in that same session or script.
 #'
 #' @return In the case of the `gsvaRanks()` method, an object of class
 #' [`gsvaRanksParam-class`].
@@ -297,7 +309,8 @@ zorder_rankstat <- function(z, p) {
 setMethod("gsvaRanks", signature(param="gsvaParam"),
           function(param,
                    verbose=TRUE,
-                   BPPARAM=SerialParam(progressbar=verbose))
+                   BPPARAM=SerialParam(progressbar=verbose),
+                   maxmem="auto")
           {
               if (verbose && gsva_global$show_start_and_end_messages) {
                   cli_alert_info(sprintf("GSVA version %s",
@@ -306,8 +319,10 @@ setMethod("gsvaRanks", signature(param="gsvaParam"),
 
               exprData <- get_exprData(param)
               dataMatrix <- unwrapData(exprData, get_assay(param))
+              maxmem <- .check_maxmem(maxmem, verbose)
+              ondisk <- .check_ondisk(param, maxmem, verbose)
 
-              if (is(dataMatrix, "DelayedMatrix") && get_ondisk(param) == "no") {
+              if (is(dataMatrix, "DelayedMatrix") && ondisk == "no") {
                   if (verbose)
                       cli_alert_info("Loading input expression data into main memory")
                   if (is_sparse(dataMatrix)) {
@@ -319,25 +334,32 @@ setMethod("gsvaRanks", signature(param="gsvaParam"),
                       dataMatrix <- as.matrix(dataMatrix)
               }
 
-              filteredDataMatrix <- .filterGenes(dataMatrix, removeConstant=TRUE,
-                                                 removeNzConstant=TRUE, verbose,
-                                                 BPPARAM=BPPARAM)
+              filtDataMatrix <- dataMatrix
+              if (get_filterRows(param))
+                  filtDataMatrix <- .filterGenes(dataMatrix, anyNA(param),
+                                                 removeConstant=TRUE,
+                                                 removeNzConstant=TRUE,
+                                                 verbose, BPPARAM=BPPARAM,
+                                                 maxmem=maxmem)
+              else if (verbose)
+                  cli_alert_warning("Skipping filtering of constant rows (filterRows=FALSE)")
               
               if (verbose)
                   cli_alert_info(sprintf("Calculating GSVA ranks"))
 
               kcdfminssize <-get_kcdfNoneMinSampleSize(param)
-              gsvarnks <- .compute_gsva_ranks(expr=filteredDataMatrix,
+              gsvarnks <- .compute_gsva_ranks(expr=filtDataMatrix,
                                               kcdf=get_kcdf(param),
                                               kcdf.min.ssize=kcdfminssize,
                                               sparse=get_sparse(param),
                                               any_na=anyNA(param),
                                               na_use=get_NAuse(param),
                                               verbose=verbose,
-                                              BPPARAM=BPPARAM)
+                                              BPPARAM=BPPARAM,
+                                              maxmem=maxmem)
 
-              rownames(gsvarnks) <- rownames(filteredDataMatrix)
-              colnames(gsvarnks) <- colnames(filteredDataMatrix)
+              rownames(gsvarnks) <- rownames(filtDataMatrix)
+              colnames(gsvarnks) <- colnames(filtDataMatrix)
 
               rnkcontainer <- wrapData(get_exprData(param), gsvarnks)
               rval <- new("gsvaRanksParam",
@@ -350,8 +372,8 @@ setMethod("gsvaRanks", signature(param="gsvaParam"),
                           absRanking=get_absRanking(param),
                           sparse=get_sparse(param), checkNA=get_checkNA(param),
                           didCheckNA=get_didCheckNA(param), anyNA=anyNA(param),
-                          use=get_NAuse(param), nzcount=nzcount(param),
-                          ondisk=get_ondisk(param))
+                          use=get_NAuse(param), filterRows=get_filterRows(param),
+                          nzcount=nzcount(param), ondisk=get_ondisk(param))
 
               if (verbose && gsva_global$show_start_and_end_messages)
                   cli_alert_success("Calculations finished")
@@ -443,8 +465,8 @@ setMethod("gsvaRanks", signature(param="gsvaParam"),
 #' @exportMethod gsvaScores
 setMethod("gsvaScores", signature(param="gsvaRanksParam"),
           function(param, verbose=TRUE,
-                   BPPARAM=SerialParam(progressbar=verbose))
-          {
+                   BPPARAM=SerialParam(progressbar=verbose),
+                   maxmem="auto") {
               if (verbose && gsva_global$show_start_and_end_messages) {
                   cli_alert_info(sprintf("GSVA version %s",
                                          packageDescription("GSVA")[["Version"]]))
@@ -452,13 +474,14 @@ setMethod("gsvaScores", signature(param="gsvaRanksParam"),
 
               ## assuming rows in the rank data have been already filtered
               exprData <- get_exprData(param)
-              filteredDataMatrix <- unwrapData(exprData, get_assay(param))
-              filteredMappedGeneSets <- .filterAndMapGeneSets(param=param,
-                                                              filteredDataMatrix=filteredDataMatrix,
-                                                              verbose=verbose)
+              filtDataMatrix <- unwrapData(exprData, get_assay(param))
+
+              filtMappedGeneSets <- .filterAndMapGeneSets(param=param,
+                                           filteredDataMatrix=filtDataMatrix,
+                                           verbose=verbose)
 
               sparse <- get_sparse(param)
-              if (sparse && !is_sparse(filteredDataMatrix))
+              if (sparse && !is_sparse(filtDataMatrix))
                   sparse <- FALSE
 
               if (verbose) {
@@ -468,8 +491,23 @@ setMethod("gsvaScores", signature(param="gsvaRanksParam"),
                     cli_alert_info("GSVA dense (classical) algorithm")
               }
 
-              if (bpnworkers(BPPARAM) > 1 && nrow(filteredDataMatrix) > 100 &&
-                  ncol(filteredDataMatrix) > 100) {
+              maxmem <- .check_maxmem(maxmem, verbose)
+              ondisk <- .check_ondisk(param, maxmem, verbose)
+
+              if (is(filtDataMatrix, "DelayedMatrix") && ondisk == "no") {
+                  if (verbose)
+                      cli_alert_info("Loading input expression data into main memory")
+                  if (is_sparse(filtDataMatrix)) {
+                      if (nzcount(param) < .Machine$integer.max)
+                          filtDataMatrix <- as(filtDataMatrix, "dgCMatrix")
+                      else
+                          filtDataMatrix <- as(filtDataMatrix, "SVT_SparseArray")
+                  } else
+                      filtDataMatrix <- as.matrix(filtDataMatrix)
+              }
+
+              if (bpnworkers(BPPARAM) > 1 && nrow(filtDataMatrix) > 100 &&
+                  ncol(filtDataMatrix) > 100) {
                   if (verbose) {
                       msg <- sprintf("Calculating GSVA scores with %d cores",
                                      as.integer(bpnworkers(BPPARAM)))
@@ -481,23 +519,37 @@ setMethod("gsvaScores", signature(param="gsvaRanksParam"),
                   BPPARAM <- NULL
               }
 
-              gsva_es <- .processMatrixCols(filteredDataMatrix,
+              ondisk <- FALSE
+              esreqmem <- as.numeric(length(filtMappedGeneSets)) *
+                          as.numeric(ncol(filtDataMatrix)) * 8 ## 8 bytes per double
+              if (esreqmem > maxmem) {
+                cli_alert_warning("The resulting matrix of enrichment scores will not fit")
+                cli_alert_warning("in the given maximum main memory size, the returned")
+                cli_alert_warning("object will use an on-disk data structure")
+                ondisk <- TRUE
+              }
+
+              gsva_es <- .processMatrixCols(filtDataMatrix,
                                             FUN=.compute_gsva_scores,
-                                            geneSetsIdx=filteredMappedGeneSets,
+                                            geneSetsIdx=filtMappedGeneSets,
                                             tau=get_tau(param),
                                             maxDiff=get_maxDiff(param),
                                             absRanking=get_absRanking(param),
                                             sparse=sparse, any_na=anyNA(param),
                                             na_use=get_NAuse(param),
                                             minSize=get_minSize(param),
-                                            verbose=verbose, minparrows=100,
-                                            minparcols=100, BPPARAM=BPPARAM)
+                                            ondisk=ondisk, verbose=verbose,
+                                            minparrows=100, minparcols=100,
+                                            BPPARAM=BPPARAM,
+                                            maxmem=ceiling(maxmem/100)) ## use
+                                            ## of memory increases here about
+                                            ## 10-fold over block size memory
 
-              rownames(gsva_es) <- names(filteredMappedGeneSets)
-              colnames(gsva_es) <- colnames(filteredDataMatrix)
+              rownames(gsva_es) <- names(filtMappedGeneSets)
+              colnames(gsva_es) <- colnames(filtDataMatrix)
 
-              gs <- .geneSetsIndices2Names(indices=filteredMappedGeneSets,
-                                           names=rownames(filteredDataMatrix))
+              gs <- .geneSetsIndices2Names(indices=filtMappedGeneSets,
+                                           names=rownames(filtDataMatrix))
               rval <- wrapData(get_exprData(param), gsva_es, gs)
 
               if (verbose && gsva_global$show_start_and_end_messages)
@@ -616,15 +668,16 @@ setMethod("gsvaEnrichment", signature(param="gsvaRanksParam"),
               minsize <- get_minSize(param)
 
               exprData <- get_exprData(param)
-              filteredDataMatrix <- unwrapData(exprData, get_assay(param))
+              filtDataMatrix <- unwrapData(exprData, get_assay(param))
 
               ## no need for verbosity when mapping a single gene set
-              filteredMappedGeneSets <- .filterAndMapGeneSets(param, wgset=geneSet,
-                                                              filteredDataMatrix,
-                                                              verbose=FALSE)
+              filtMappedGeneSets <- .filterAndMapGeneSets(param,
+                                           wgset=geneSet,
+                                           filteredDataMatrix=filtDataMatrix,
+                                           verbose=FALSE)
 
-              geneSetIdx <- filteredMappedGeneSets[[1]]
-              edata <- .gsva_enrichment_data(R=filteredDataMatrix,
+              geneSetIdx <- filtMappedGeneSets[[1]]
+              edata <- .gsva_enrichment_data(R=filtDataMatrix,
                                              column=column,
                                              geneSetIdx=geneSetIdx,
                                              maxDiff=maxDiff,
@@ -705,7 +758,7 @@ setMethod("gsvaEnrichment", signature(param="gsvaRanksParam"),
 #' @importFrom sparseMatrixStats colRanks
 .compute_gsva_ranks <- function(expr, kcdf, kcdf.min.ssize,
                                 sparse, any_na, na_use, verbose,
-                                BPPARAM=NULL) {
+                                BPPARAM=NULL, maxmem=Inf) {
 
     kcdfparam <- .parse_kcdf_param(expr, kcdf, kcdf.min.ssize, sparse, verbose)
     kernel <- kcdfparam$kernel
@@ -726,7 +779,7 @@ setMethod("gsvaEnrichment", signature(param="gsvaRanksParam"),
     Z <- .processMatrixRows(expr, FUN=compute.gene.cdf, Gaussk=Gaussk,
                             kernel=kernel, sparse=sparse, any_na=any_na,
                             na_use=na_use, verbose=verbose, minparrows=100,
-                            minparcols=100, BPPARAM=BPPARAM)
+                            minparcols=100, BPPARAM=BPPARAM, maxmem=maxmem)
 
     if (!is.null(BPPARAM) && verbose) {
         msg <- sprintf("Calculating column ranks with %d cores",
@@ -739,7 +792,7 @@ setMethod("gsvaEnrichment", signature(param="gsvaRanksParam"),
     ## from 'order()' based on ranks
     R <- .processMatrixCols(Z, FUN=compute.col.ranks, ties.method="last",
                             verbose=verbose, minparrows=100, minparcols=100,
-                            BPPARAM=BPPARAM)
+                            BPPARAM=BPPARAM, maxmem=maxmem)
 
     return(R)
 }
@@ -931,9 +984,10 @@ setMethod("gsvaEnrichment", signature(param="gsvaRanksParam"),
 
     if (all(na_mask))
         return(list(dos=matrix(NA_integer_, nrow(r), ncol(r)),
-                    srs=matrix(NA, nrow(r), ncol(r))))
+                    srs=matrix(NA_real_, nrow(r), ncol(r))))
 
     n_nas <- colSums(na_mask)
+    mode(n_nas) <- "integer"
     mask <- !na_mask & mask
     p <- nrow(r)
     r_dense <- as.matrix(r)
@@ -976,12 +1030,14 @@ setMethod("gsvaEnrichment", signature(param="gsvaRanksParam"),
         return(list(dos=rep(NA, length(r)), srs=rep(NA, length(r))))
 
     n_nas <- sum(na_mask)
+    mode(n_nas) <- "integer"
     mask <- !na_mask & r == 0
     p <- length(r)
     r_dense <- as.integer(r)          ## assume ranks are integer
 
     if (any(mask)) {                  ## sparse ranks into dense ranks
         nzs <- sum(mask)
+        mode(nzs) <- "integer"
         r_dense[!mask] <- r_dense[!mask] + nzs ## shift ranks of nonzero values
         r_dense[mask] <- seq.int(nzs)          ## zeros get increasing ranks
     }
@@ -1004,8 +1060,8 @@ setMethod("gsvaEnrichment", signature(param="gsvaRanksParam"),
 #' @importFrom BiocParallel bpnworkers
 #' @importFrom S4Arrays is_sparse refdim
 .compute_gsva_scores <- function(R, geneSetsIdx, tau, maxDiff, absRanking,
-                                 sparse, any_na, na_use, minSize,
-                                 grid=NULL, grid_es=NULL, verbose) {
+                                 sparse, any_na, na_use, minSize, ondisk,
+                                 verbose) {
     p <- nrow(R)
     n <- ncol(R)
     es <- NULL
@@ -1015,18 +1071,16 @@ setMethod("gsvaEnrichment", signature(param="gsvaRanksParam"),
     wna_env <- new.env()
     assign("w", FALSE, envir=wna_env)
     es <- NULL
-    if (is(R, "DelayedMatrix")) {
+    if (is(R, "DelayedMatrix") || ondisk) {
         sink <- HDF5RealizationSink(c(length(geneSetsIdx), ncol(R)),
                                     as.sparse=FALSE) ## GSVA scores are dense
-        if (is.null(grid))
-            grid <- DummyArrayGrid(dim(R))
-        if (is.null(grid_es))
-            grid_es <- DummyArrayGrid(dim(sink))
+        grid <- DummyArrayGrid(dim(R))
+        grid_es <- DummyArrayGrid(dim(sink))
 
         if (length(grid) != length(grid_es) ||
             refdim(grid)[2] != refdim(grid_es)[2] ||
             dim(grid)[2] != dim(grid_es)[2]) {
-            cli_abort("x"="Grid column blocks for ranks should match grid column blocks for enrichment scores")
+            cli_abort(c("x"="Grid column blocks for ranks should match grid column blocks for enrichment scores"))
         }
 
         ## avp - ArrayViewport for reaching the (possibly sparse) rank matrix
@@ -1043,7 +1097,7 @@ setMethod("gsvaEnrichment", signature(param="gsvaRanksParam"),
                                           symRnkStat=rnkstats$srs,
                                           maxDiff, absRanking, tau,
                                           any_na, na_use, minSize,
-                                          wna_env, verbose=FALSE)
+                                          wna_env, verbose=verbose)
             write_block(sink, avp_es, block)
         }
 
@@ -1064,7 +1118,7 @@ setMethod("gsvaEnrichment", signature(param="gsvaRanksParam"),
                                     decOrdStat=rnkstats$dos,
                                     symRnkStat=rnkstats$srs,
                                     maxDiff, absRanking, tau, any_na,
-                                    na_use, minSize, wna_env, verbose=FALSE)
+                                    na_use, minSize, wna_env, verbose=verbose)
     }
 
     if (any_na && na_use == "na.rm")
@@ -1245,9 +1299,15 @@ setMethod("gsvaEnrichment", signature(param="gsvaRanksParam"),
 ## functions interfacing C code
 ##
 
-.fetch_row_nzvals <- function(X, i) {
+.fetch_row_nzvals <- function(X, i, whimin1=NULL) {
   stopifnot(is(X, "SVT_SparseArray")) ## QC
-  .Call("fetch_row_nzvals_R", X, as.integer(i))
+  stopifnot(is.numeric(i)) ## QC
+  if (!is.null(whimin1)) {
+      stopifnot(is.numeric(whimin1)) ## QC
+      whimin1 <- as.integer(whimin1)
+      stopifnot(length(whimin1) == ncol(X)) ## QC
+  }
+  .Call("fetch_row_nzvals_R", X, as.integer(i), whimin1)
 }
 
 .ecdfvals_svt_to_dense <- function(X, verbose) {
@@ -1269,7 +1329,8 @@ setMethod("gsvaEnrichment", signature(param="gsvaRanksParam"),
 }
 
 #' @importFrom HDF5Array HDF5RealizationSink
-#' @importFrom DelayedArray seed DummyArrayGrid gridReduce
+#' @importFrom S4Arrays DummyArrayGrid
+#' @importFrom DelayedArray seed gridReduce
 .ecdfvals_sparseh5_to_sparseh5 <- function(X, grid=NULL, verbose=FALSE) {
   stopifnot(is(X, "DelayedMatrix") || is(X, "HDF5Matrix")) ## QC
 
@@ -1289,7 +1350,8 @@ setMethod("gsvaEnrichment", signature(param="gsvaRanksParam"),
 }
 
 #' @importFrom HDF5Array HDF5RealizationSink
-#' @importFrom DelayedArray seed DummyArrayGrid gridReduce
+#' @importFrom S4Arrays DummyArrayGrid
+#' @importFrom DelayedArray seed gridReduce
 .ecdfvals_sparseh5_to_denseh5 <- function(X, grid=NULL, verbose) {
   stopifnot(is(X, "DelayedMatrix") || is(X, "HDF5Matrix")) ## QC
 
@@ -1491,8 +1553,8 @@ setMethod("gsvaEnrichment", signature(param="gsvaRanksParam"),
     if (is.null(dim(symRnkStat)))
       symRnkStat <- matrix(symRnkStat, ncol=1)
   } else {
-    decOrdStat <- decOrdStat[, colIdx]
-    symRnkStat <- symRnkStat[, colIdx]
+    decOrdStat <- decOrdStat[, colIdx, drop=FALSE]
+    symRnkStat <- symRnkStat[, colIdx, drop=FALSE]
   }
   sco <- .Call("gsva_score_genesets_R", geneSetsIdx, decOrdStat, symRnkStat,
                maxDiff, absRanking, as.double(tau), any_na, na_use, minSize,
@@ -1529,28 +1591,6 @@ setMethod("gsvaEnrichment", signature(param="gsvaRanksParam"),
 
 ## calculate ranks using an HDF5 backend
 
-## adapted from .define_multiworker_grid() in beachmat/R/colBlockApply.R
-#' @importFrom DelayedArray rowAutoGrid colAutoGrid getAutoBlockLength type
-rowgridsize <- function(X, nworkers) {
-  max.block.length <- getAutoBlockLength(type(X))
-  if (nworkers > 1)
-      max.block.length <- .Machine$integer.max
-  expected.block.length <- max(1, ceiling(nrow(X) / nworkers) * ncol(X))
-  block.length <- min(max.block.length, expected.block.length)
-  rowAutoGrid(X, block.length=block.length)
-}
-
-## adapted from .define_multiworker_grid() in beachmat/R/colBlockApply.R
-#' @importFrom DelayedArray rowAutoGrid colAutoGrid getAutoBlockLength type
-colgridsize <- function(X, nworkers) {
-  max.block.length <- getAutoBlockLength(type(X))
-  if (nworkers > 1)
-      max.block.length <- .Machine$integer.max
-  expected.block.length <- max(1, ceiling(ncol(X) / nworkers) * nrow(X))
-  block.length <- min(max.block.length, expected.block.length)
-  colAutoGrid(X, block.length=block.length)
-}
-
 #' @importFrom MatrixGenerics colRanks
 #' @importFrom BiocParallel SerialParam
 .colRanksHDF5 <- function(X, grid=NULL, ties.method="last") {
