@@ -113,6 +113,16 @@
 #' happens, and giving an error if no values are left after removing the `NA`
 #' values.
 #'
+#' @param filterRows Logical vector of length 1, indicating whether the rows in,
+#' the input expression data, typically corresponding to transcripts, genes or
+#' proteins, should be filtered for constant expression across columns,
+#' typically corresponding to samples or cells, with respect to all available
+#' (nonmissing) values and to the non-zero values. By default, this slot is set
+#' to `TRUE` and the user may set it to `FALSE` when there is absolute certainty
+#' that no such rows exist in the input expression data, since this may save
+#' running time, especially with data sets with hundreds of thousands or
+#' millions of columns.
+#'
 #' @param ondisk Character vector of length 1 denoting whether an on-disk backend
 #' should be used to reduce the memory footprint. The default value
 #' `ondisk="auto"` will attempt to load all the data in main memory when the
@@ -146,20 +156,24 @@
 #' \doi{10.1186/1471-2105-14-7}
 #'
 #' @examples
+#' suppressPackageStartupMessages({
+#' library(GSEABase)
 #' library(GSVA)
-#' suppressPackageStartupMessages(library(GSVAdata))
+#' library(GSVAdata)
+#' })
 #'
-#' data(leukemia)
+#' data(geneprotExpCostaEtAl2021)
 #' data(c2BroadSets)
 #' 
 #' ## for simplicity, use only a subset of the sample data
-#' ses <- leukemia_eset[1:1000, ]
+#' se <- geneExpCostaEtAl2021[1:1000, ]
 #' gsc <- c2BroadSets[1:100]
-#' gp1 <- gsvaParam(ses, gsc)
+#' gp1 <- gsvaParam(se, gsc)
 #' gp1
 #'
 #'
 #' @importFrom methods new
+#' @importFrom cli cli_alert_warning
 #' @rdname gsvaParam-class
 #' 
 #' @export
@@ -171,6 +185,7 @@ gsvaParam <- function(exprData, geneSets,
                       absRanking=FALSE, sparse=TRUE,
                       checkNA=c("auto", "yes", "no"),
                       use=c("everything", "all.obs", "na.rm"),
+                      filterRows=TRUE,
                       ondisk=c("auto", "yes", "no"),
                       verbose=TRUE) {
     kcdf <- match.arg(kcdf)
@@ -206,24 +221,29 @@ gsvaParam <- function(exprData, geneSets,
                                     checkNA=checkNA, use=use)
 
     nzc <- .estimate_nzcount(exprData, assay, verbose)
-    if (ondisk == "auto") {
-      if (nzc >= .Machine$integer.max) {
-        ondisk <- "yes"
-      } else
-        ondisk <- "no"
-    }
 
     if (!is_sparse(unwrapData(exprData, assay))) ## use sparse regime only
         sparse <- FALSE                          ## when input is sparse
 
-    new("gsvaParam",
-        exprData=exprData, geneSets=geneSets,
-        assay=assay, annotation=annotation,
-        minSize=minSize, maxSize=maxSize,
-        kcdf=kcdf, kcdfNoneMinSampleSize=kcdfNoneMinSampleSize,
-        tau=as.double(tau), maxDiff=maxDiff, absRanking=absRanking,
-        sparse=sparse, checkNA=checkNA, didCheckNA=naparam$didCheckNA,
-        anyNA=naparam$any_na, use=use, nzcount=nzc, ondisk=ondisk)
+    if (!filterRows) {
+        cli_alert_warning("filterRows=FALSE and rows with constant values will not be filtered out")
+        cli_alert_warning("Use it only if you are sure that such rows are not present in the input data")
+    }
+
+    param <- new("gsvaParam",
+                 exprData=exprData, geneSets=geneSets,
+                 assay=assay, annotation=annotation,
+                 minSize=minSize, maxSize=maxSize,
+                 kcdf=kcdf, kcdfNoneMinSampleSize=kcdfNoneMinSampleSize,
+                 tau=as.double(tau), maxDiff=maxDiff, absRanking=absRanking,
+                 sparse=sparse, checkNA=checkNA, didCheckNA=naparam$didCheckNA,
+                 anyNA=naparam$any_na, use=use, filterRows=filterRows,
+                 nzcount=nzc, ondisk=ondisk)
+
+    maxmem <- .check_maxmem("auto", verbose)
+    .check_ondisk(param, maxmem, verbose)
+
+    return(param)
 }
 
 
@@ -320,6 +340,12 @@ setValidity("gsvaParam", function(object) {
     if(!.isCharLength1(object@use)) {
         inv <- c(inv, "@use must be a single character string")
     }
+    if(length(object@filterRows) != 1) {
+        inv <- c(inv, "@filterRows must be of length 1")
+    }
+    if(is.na(object@filterRows)) {
+        inv <- c(inv, "@filterRows must not be NA")
+    }
     if(length(object@nzcount) != 1) {
         inv <- c(inv, "@nzcount must be of length 1")
     }
@@ -368,6 +394,12 @@ get_sparse <- function(object) {
   return(object@sparse)
 }
 
+#' @noRd
+get_filterRows <- function(object) {
+  stopifnot(inherits(object, "gsvaParam"))
+  return(object@filterRows)
+}
+
 #' @noRd
 get_ondisk <- function(object) {
   stopifnot(inherits(object, "gsvaParam"))
@@ -382,8 +414,8 @@ get_ondisk <- function(object) {
 #' @param recursive Not used with `x` being an object of
 #' class [`gsvaParam-class`].
 #'
-#' @aliases anyNA,ssgseaParam-method
-#' @rdname ssgseaParam-class
+#' @aliases anyNA,gsvaParam-method
+#' @rdname gsvaParam-class
 setMethod("anyNA", signature=c("gsvaParam"),
           function(x, recursive=FALSE)
             return(x@anyNA))
@@ -417,6 +449,7 @@ setMethod("show",
                       cat("missing data: no\n")
               } else
                   cat("missing data: didn't check\n")
+              cat("filterRows: ", get_filterRows(object), "\n")
               nzcmsg <- sprintf("nonzero values: %s than 2^31 (INT_MAX)\n",
                                 ifelse(nzcount(object) > .Machine$integer.max,
                                        "more", "less"))
